// The MIT License

// Copyright (c) 2010 Massachusetts Institute of Technology

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

// Author: Richard Uhler ruhler@mit.edu
import ClientServer::*;
import FixedPoint::*;
import FIFO::*;
import GetPut::*;

typedef Server#(
    Tuple2#(word, word),
    Tuple2#(word, word)
) Divider#(type word);

// Unsigned division
// Input: n, d
// Output: q, r
//  where n = d*q + r.
//
// This implementation uses the non-restoring algorithm (or so I'm told).
// itersPerCycle is an integer specifying how many iterations of the algorithm
// should be performed each clock cycle. This number should divide the bit
// width evenly.
module mkNonRestoringDivider(Integer itersPerCycle, Divider#(Bit#(ws)) ifc)
    provisos(Add#(_, 1, ws));

    if (valueof(ws) % itersPerCycle != 0) begin
        error("itersPerCycle must evenly divide operand bit width");
    end

    FIFO#(Tuple2#(Bit#(ws), Bit#(ws))) incoming <- mkFIFO();
    FIFO#(Tuple2#(Bit#(ws), Bit#(ws))) outgoing <- mkFIFO();

    Reg#(Bool) busy <- mkReg(False);
    Reg#(Bit#(ws)) xReg <- mkRegU();
    Reg#(Bit#(ws)) dReg <- mkRegU();
    Reg#(Bit#(ws)) pReg <- mkRegU();
    Reg#(Bit#(TAdd#(1, TLog#(ws)))) iReg <- mkRegU();

    rule start (!busy);
        match {.ix, .id} <- toGet(incoming).get();

        busy <= True;
        dReg <= id;
        xReg <= ix;
        pReg <= 0;
        iReg <= 0;
    endrule

    // Return the most significant bit of a bit vector
    function Bit#(1) top(Bit#(n) x) = x[valueof(n)-1];

    // Return all but the most significant bit of a bit vector
    function Bit#(TSub#(n,1)) rest(Bit#(n) x) = x[valueof(n)-2:0];

    // Returns new x, p after a single iteration of the division algorithm.
    function Tuple2#(Bit#(ws), Bit#(ws)) iterate(Bit#(ws) x, Bit#(ws) p, Bit#(ws) d);
        if (top(p) == 1) begin
            p = ((p << 1) | zeroExtend(top(x))) + d;
        end else begin
            p = ((p << 1) | zeroExtend(top(x))) - d;
        end
        x = (x << 1) | zeroExtend(~top(p));
        return tuple2(x, p);
    endfunction

    rule doiterate (busy);
        Bit#(ws) x = xReg;
        Bit#(ws) p = pReg;

        for (Integer i = 0; i < itersPerCycle; i = i+1) begin
            let iout = iterate(x, p, dReg);
            x = tpl_1(iout);
            p = tpl_2(iout);
        end

        if (iReg + fromInteger(itersPerCycle) == fromInteger(valueof(ws))) begin
            if (top(p) == 1) begin
                p = p + dReg;
            end
            outgoing.enq(tuple2(x, p));
            busy <= False;
        end else begin
            xReg <= x;
            pReg <= p;
            iReg <= iReg + fromInteger(itersPerCycle);
        end
    endrule
        
    interface Put request = toPut(incoming);
    interface Get response = toGet(outgoing);

endmodule

// Fixed Point divider
module mkFixedPointDivider(Integer itersPerCycle, Divider#(FixedPoint#(iw, fw)) ifc)
    provisos(
        Add#(a__, TAdd#(iw, fw), TAdd#(iw, TMul#(2, fw))),
        Add#(b__, 1, TAdd#(iw, TMul#(2, fw)))
    );


    // We use the integer division algorithm to do fixed point division.
    //
    // If you pack a fixed point number x,
    // you get a new number which is x * 2^fw
    // If you unpack a number y into a fixed point number, you get a fixed point
    // number which is y / 2^fw.
    //
    // Let the inputs be two fixed point numbers a, b. We want to generate the
    // fixed point number a / b.
    // We unpack both numbers, multiply the numerator by 2^fw, perform integer
    // division and get the integer:
    // (a * 2^fw) * 2^fw / (b & 2^fw) = (a/b) * 2^fw
    // Simply unpack and we get our fixed point a/b.

    Divider#(Bit#(TAdd#(iw, TMul#(2, fw)))) div <- mkNonRestoringDivider(itersPerCycle);
    FIFO#(Bool) negate <- mkFIFO();

    interface Put request;
        method Action put(Tuple2#(FixedPoint#(iw, fw), FixedPoint#(iw, fw)) x);
            match {.a, .b} = x;
            Bool neg = False;

            if (pack(fxptGetInt(a))[valueof(iw)-1] == 1'b1) begin
                neg = !neg;
                a = -a;
            end

            if (pack(fxptGetInt(b))[valueof(iw)-1] == 1'b1) begin
                neg = !neg;
                b = -b;
            end
            negate.enq(neg);

            Bit#(TAdd#(iw, fw)) aa = pack(a);
            Bit#(TAdd#(iw, TMul#(2, fw))) aaa = zeroExtend(aa);
            aaa = aaa << valueof(fw);

            Bit#(TAdd#(iw, fw)) bb = pack(b);
            Bit#(TAdd#(iw, TMul#(2, fw))) bbb = zeroExtend(bb);
            
            div.request.put(tuple2(aaa, bbb));
        endmethod
    endinterface

    interface Get response;
        method ActionValue#(Tuple2#(FixedPoint#(iw, fw), FixedPoint#(iw, fw))) get();
            match {.q, .r} <- div.response.get();
            Bit#(TAdd#(iw, fw)) qb = q[valueof(iw) + valueof(fw)-1:0];
            FixedPoint#(iw, fw) qf = unpack(qb);

            if (negate.first()) begin
                qf = -qf;
            end

            negate.deq();


            // TODO: return a useful remainder
            return tuple2(qf, ?); 
        endmethod
    endinterface

endmodule

